/***************************************************************************

  paclearn.c

  Let the computer learn how to play PacMan!
  Learns from its own mistakes; will eventually find the BEST pattern
  possible.

  Uses the search-tree structures defined in pactree.c
  and the analysis engine implemented in pacthink.c
  
  Module added by Sean Gugler on 97-08-28
  modified 97-09-10 Sean Gugler - version 0.8
  modified 97-10-01 Sean Gugler - version 0.9

***************************************************************************/

#include "pacdefs.h"
#include "pactree.h"

#include <stdio.h>
#include <stdlib.h>
const char * const dTXT[16] = {"X","U","L","UL","R","UR","LR","ULR","D","UD","LD","ULD","RD","URD","LRD","ULRD"};

int maskOptions = 0x0f;

enum
{
    eLearnMode_play,
    eLearnMode_exhaust,
    eLearnMode_explore
};
int modeLearn;

typedef struct s_LearnState {
    LearnPath path;
    LearnTable table;
    int nLevel;
    LearnPath pathExplore;
    int bBest;
} LearnState;

static LearnState ls;

static LearnMove movesKillMe[] = {4, 4, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0};
static struct s_LearnPath s_pathKillMe = {movesKillMe, movesKillMe, 0, 0};
static LearnPath pathKillMe = &s_pathKillMe;

static const char * textGame = "PACMAN";
static const char fmt[2][16] = { "%s.E%02d", "%s.B%02d" };

/* helpers */
static void table_write_delete();
static int  table_read();
static void table_new();
static void table_add_path();

/* pactape.c */
extern int  pactape_get_dsw   ();
extern int  pactape_check_dsw (int dswCheck);

/* pacthink.c */
extern int  pacthink_choose (int dirs);
extern long pacthink_score  ();

/***************************************************************************

  Initialization

***************************************************************************/
void paclearn_init (const char * name)
{
    textGame = name;
    if (ls.pathExplore == 0)
        ls.pathExplore = LearnPath_new (-1);
}

/***************************************************************************

  Called on the start of every level

***************************************************************************/
int paclearn_start (int nLevel, int bBestSolved)
{
    if (ls.nLevel != nLevel      ||
        ls.bBest  != bBestSolved )
    {
        if (ls.table != tableNull)
            table_write_delete();
        ls.nLevel = nLevel;
        ls.bBest  = bBestSolved;
        if (!table_read())
        {
            if (bBestSolved)
                return 0;
            else
                table_new();
        }
        else if (ls.table->nData > 0
                 && !pactape_check_dsw(ls.table->data[0]))
            return 0;
    }
    ls.path = LearnTable_best (ls.table);
    modeLearn = eLearnMode_play;
    LearnPath_rewind (ls.path);
    return 1;
}

/***************************************************************************

  Request a move from:
   - playback of a partially-explored path until exhausted, then
   - the pacthink analysis engine

***************************************************************************/
int paclearn_request (int options)
{
    LearnMove move = 0;
    switch (modeLearn)
    {
    case eLearnMode_play:
        {
            if (!LearnPath_end_b (ls.path))
            {
                move = LearnPath_play (ls.path);
                options = dX;
#ifdef PT_DEBUG
                sprintf (debugText[0], "PLAYING-%s.", dTXT[move]);
                debug_msg (debugText[0]);
#endif
            }
            else
            {
                int choices = LearnPath_get_choices(ls.path);
                if (choices != dX) options = choices;
                if (options & (options-1)) /* more than 1 choice left? */
                    move = pacthink_choose(options);
                else
                    move = options;
                modeLearn = eLearnMode_exhaust;
#ifdef PT_DEBUG
                sprintf (debugText[0], "EXHAUSTING-%s.", dTXT[options]);
                debug_msg (debugText[0]);
#endif
            }
            break;
        }
    case eLearnMode_explore:
        {
#ifdef PT_DEBUG
            sprintf (debugText[0], "DECIDING-%s.", dTXT[options]);
            debug_msg (debugText[0]);
#endif
            LearnPath_set_choices (ls.path, options);
            if (options & (options-1)) /* more than 1 choice left? */
                move = pacthink_choose(options);
            else
                move = options;
            break;
        }
    }
    maskOptions = (~options) & 0x0f;
    return move;
}

/***************************************************************************

  Commit to a move.
  Implemented as a separate function to allow user-override with joystick.

***************************************************************************/
void paclearn_commit (LearnMove move)
{
    if (modeLearn == eLearnMode_play)
        return;

#ifdef PT_DEBUG
    sprintf (debugText[0], "COMMIT-%s.", dTXT[move]);
    debug_msg (debugText[0]);
#endif

    maskOptions = 0x0f;
    switch (modeLearn)
    {
    case eLearnMode_exhaust:
        if (!LearnPath_end_b (ls.path))
            debug_msg ("BAD-EXHAUST.");
        {
            LearnPath_copy (ls.pathExplore, ls.path);

            LearnPath_set_choices (ls.path, LearnPath_get_choices (ls.path) & (~move));
            if (LearnPath_exhausted_b (ls.path))
            {
                LearnTable_remove (ls.table, ls.path);
                LearnPath_delete (ls.path);
            }
            ls.path = ls.pathExplore;
        }
        LearnPath_add (ls.path, move);
        modeLearn = eLearnMode_explore;
        break;

    case eLearnMode_explore:
        LearnPath_set_choices (ls.path, LearnPath_get_choices (ls.path) & (~move));
        if (!LearnPath_exhausted_b (ls.path))
            table_add_path();
        LearnPath_add (ls.path, move);
        break;
    }
}

/***************************************************************************

  Called whenever input ceases:
    - at the end of every level, and
    - when pac-man dies

***************************************************************************/
void paclearn_end (int bSurvived)
{
    if (bSurvived)
    {
        if (ls.bBest)
            ls.path = pathKillMe;
        else
        {   /* Swap out the exploration table and
               swap in the solution table,
               so we can store the winning path there. */
            table_write_delete();
            ls.bBest = 1;
            if (!table_read())
                table_new();
            table_add_path();
        }
    }
    if (ls.path != pathKillMe)
    {
        if (modeLearn == eLearnMode_play)
        {
            LearnTable_remove (ls.table, ls.path);
            LearnPath_delete (ls.path);
        }
        ls.path = pathKillMe;
    }
    modeLearn = eLearnMode_play;
    LearnPath_rewind (ls.path);
}

/***************************************************************************

  Clean-up, flush to storage

***************************************************************************/
void paclearn_shutdown ()
{
    if (ls.table != tableNull)
    {
        if (modeLearn == eLearnMode_explore)
        {
            LearnPath_set_choices (ls.path, dX);
            table_add_path();
        }
        table_write_delete();
    }
    if (ls.pathExplore)
    {
        LearnPath_delete (ls.pathExplore);
        ls.pathExplore = pathNull;
    }
    ls.nLevel = 0;
    debug_msg ("SHUTDOWN");
}

/***************************************************************************

  Helper functions for table manipulation

***************************************************************************/
void table_write_delete ()
{
    char filename [16];
    if (ls.nLevel == 0) debug_msg ("WRITE-0");
    sprintf (filename, fmt[ls.bBest], textGame, ls.nLevel);
    LearnTable_write (ls.table, filename);
    LearnTable_delete (ls.table);
    ls.table = tableNull;
}
int table_read ()
{
    char filename [16];
    sprintf (filename, fmt[ls.bBest], textGame, ls.nLevel);
    ls.table = LearnTable_read (filename);
    return (ls.table != tableNull);
}
void table_new ()
{
    ls.table = LearnTable_new ();
    ls.table->nData = 1;
    ls.table->data = (unsigned char *) malloc (1);
    ls.table->data[0] = pactape_get_dsw();
}
void table_add_path ()
{
    long nPoints = pacthink_score();
    LearnPath pathCopy = LearnPath_new (LearnPath_size (ls.path));
    LearnPath_copy (pathCopy, ls.path);
    LearnPath_set_score (pathCopy, nPoints);
    LearnTable_add (ls.table, pathCopy);
}
