/***************************************************************************

  pactape.c

  Provides the ability to record and playback Pac-Man sessions.
  Also interfaces with the Pac-Learn engine.

  Created by Sean Gugler on 97-08-21
  version 0.7  97-08-27  Sean Gugler
  version 0.8  97-09-10  Sean Gugler
  version 0.9  97-10-01  Sean Gugler
  version 1.0  97-10-03  Sean Gugler

***************************************************************************/

#include "driver.h"
#include "vidhrdw/generic.h"
#include <time.h>

#include "pacdefs.h"
#include "allegro.h"

/************************************
  Forward declarations of functions
*************************************/

/* Inject this function into the corresponding vidhrdw function */
void pactape_on_vh_update (struct osd_bitmap *bitmap);

    /* functions called by update */
    void pactape_read_keystate();
    void pactape_update_state();
    void pactape_update_mode();
    void pactape_add_display (struct osd_bitmap *bitmap);

/* Inject these functions into the device input stream */
int pactape_input_port_0_r (int offset);
int pactape_input_port_1_r (int offset);


/* TAPE-related functions */
int pactape_read_tape();
int pactape_write_tape();

/* LEARN-related functions */
extern void paclearn_init     (const char * name);
extern int  paclearn_start    (int nLevel, int bBest);
extern int  paclearn_request  (int options);
extern void paclearn_commit   (unsigned char move);
extern void paclearn_end      (int bSurvived);
extern void paclearn_shutdown ();

extern void pacthink_update_state();

/* Other helper functions */
long pactape_get_score();
int  pactape_get_dsw();
int  pactape_check_dsw (int dswCheck);

/*******************************
  DATA STRUCTURES and VARIABLES
********************************/

#define PacTapeVersion 2
static int bVersion1Hack = 0;

/* Pac-Tape depends on the DIP-SWITCH settings for difficulty */
/* (I formerly trapped # of lives and bonus life score,
    but realized it should be allowed to playback whatever
    possible, to allow mix-n-match files from various levels.
#define dswMask 0x7c
*/
#define dswMask 0x40

int nLevel;
int nScoreLevel;
int bBeginLevel;
const int nSpriteFruit = 12;
int nSpritePacman = 10;
int xPacman, yPacman;
int x,y;

int bDying     = 0;
int bPoweredUp = 0;
int bVanished  = 0;
/* RAM always updated one cycle before bVanished */
#define bTxVanished  bVanished ^ RAM[0x4dd1]
      /* Tx = Transition */

enum {
    mode_normal = 1,
    mode_learn,
    mode_best,
    mode_record,
    mode_playback
};
static int mode = mode_normal;
static int bCancelThisMode;

enum { /* used to track game states */
    st_test,
    st_demo,
    st_start,
    st_ready,
    st_play,
    st_die,
    st_win,
    st_unknown
};
static int state = st_unknown;
static int statePrevious = st_unknown;
#define bTxState (state != statePrevious)
      /* Tx = Transition */
static int bKeyPressed;

#if 1 /* <<<<< DEBUG: display state */
static const char* stateName [8] = {
    "TEST",
    "DEMO",
    "START",
    "READY",
    "PLAY",
    "DIE",
    "WIN",
    "???"
};
void pacdebug_show_state()
{
    struct DisplayText dt[2];
    dt[0].text = stateName[state];
    dt[0].color = DT_COLOR_YELLOW;
    dt[0].x = 8*11;
    dt[0].y = 8*2;
    dt[1].text = 0;
#if 1
    displaytext(dt,0);
#endif
}
#else
#define pacdebug_show_state()
#endif /* END DEBUG >>>>> */


/***** TAPE *****/

static int bTapeRunning;
static long nFrames;
#define MOVEBUF 0x800
static unsigned short moves[MOVEBUF];
static unsigned short * moveCurrent = moves;

/***** LEARN *****/

extern int modeLearn;
extern int maskOptions;

static int bLearnEngaged;
static int xLast, yLast;
static int maskChosen;
static int maskOverride; /* == maskChosen, unless player overrides with joystick */

const char * const mTXT[16] = {"ULRD","LRD","URD","RD","ULD","LD","UD","D","ULR","LR","UR","R","UL","L","U","X"};
#ifdef PT_DEBUG
struct DisplayText dt[6];
char debugText[6][28];
int bSingleStep = 0;
#endif /* PT_DEBUG */


/***************************************************************************

  This function is called every time a frame is shown on screen.

***************************************************************************/
void pactape_on_vh_update(struct osd_bitmap *bitmap)
{
    pactape_read_keystate();
    pactape_update_state();
    pacthink_update_state();
    pactape_update_mode();
    pactape_add_display(bitmap);

    /* These belong in update_state ...
       but since RAM is always updated one cycle
       before video is refreshed, we do a "late"
       state update here */
    bVanished  = RAM[0x4dd1];
    bPoweredUp = RAM[0x4da6];
    bDying     = RAM[0x4da5];
}

/***************************************************************************

  React to keyboard

***************************************************************************/
void pactape_read_keystate()
{
    const int keymap [4][2] = {
        {OSD_KEY_L, mode_learn} ,
        {OSD_KEY_B, mode_best} ,
        {OSD_KEY_R, mode_record} ,
        {OSD_KEY_A, mode_playback} };
    int i;
    for (i=3; i>=0; i--)
        if (osd_key_pressed(keymap[i][0]))
            break;
    if (i>=0)
    {
        if (bKeyPressed == 0)
        {
            if (mode != mode_normal)
            {
                bCancelThisMode = 1;
                if (mode == keymap[i][1])
                    bKeyPressed = 1;
            }
            else
            {
                mode = keymap[i][1];
                bKeyPressed = 1;
            }
        }
    }
#ifdef PT_DEBUG
    else if (osd_key_pressed(OSD_KEY_S))
    { if (bKeyPressed == 0) bSingleStep ^= 1;
      bKeyPressed = 1;
    }
#endif
    else bKeyPressed = 0;
}

/***************************************************************************

  "Read" the screen (video memory) to determine game state.

***************************************************************************/
void pactape_update_state()
{
    unsigned char chVid = VID(11,18);
    statePrevious = state;

    /* Version 1 checked state this way.  Tapes
       recorded with Version 1 won't playback properly
       past a death sequence if this isn't done. */
    if (bVersion1Hack) bDying = (spriteram[nSpritePacman]>>2 == 0x34);

/*
 location (11,20) ------+
                        |
                        V
      0x52 = 'R' in    "READY"
      0x53 = 'S' in    "START" (piranha only)
      0x4d = 'M' in  "GAME  OVER"
      0x0e = 'E' in "0A4E82C" (first pattern of memory check)
      0x40 = ' ' during game play
      0x3c = corner of "box" test pattern
*/
    if (chVid == 0x0e)                     {state = st_unknown;
                                            bCancelThisMode = 1;}
    else switch (state)
    {
    default:
    case st_unknown: if  (chVid == 0x3c)    state = st_test;     break;
    case st_test:    if  (chVid == 0x40)    state = st_demo;     break;
    case st_demo:    if  ( RAM[0x4dd6])    {state = st_start;
                                            srandom (uclock());} break;
    case st_die:     if  (chVid == 0x4d)    state = st_demo; /* fall through */
    case st_win:
    case st_start:   if ((!RAM[0x4dd6]) &&
                        ((chVid == 0x52) ||
                         (chVid == 0x53) )) state = st_ready;    break;
    case st_ready:   if  (chVid == 0x40)   {state = st_play;
                                            bBeginLevel = 0;}    break;
    case st_play:    if (bDying)            state = st_die;
                else if (colorram[0x3a0] == 0x1f)
                                            state = st_win;      break;
    }

    /* track PACMAN sprite:  #5 normally, #1 during power up */
    nSpritePacman = bPoweredUp ? 2 : 10;

    /* track level beginning and number */
    if (state == st_ready &&
        (statePrevious == st_start || statePrevious == st_win))
    {
        bBeginLevel = 1;
        nLevel++;
        nScoreLevel = pactape_get_score();
    }
}

/***************************************************************************

  Update mode state based on game state

***************************************************************************/
void pactape_update_mode()
{
    if (mode == mode_playback)
    {
        if (bBeginLevel && (bTxState || !bTapeRunning))
        {
            if (!pactape_read_tape())
                bCancelThisMode = 1;
            else
            {
                bTapeRunning = 1;
                moveCurrent = moves;
                nFrames = 0;
            }
        }
    }
    else if (mode == mode_record)
    {
        if (bTapeRunning)
        {
            if      (state == st_win && bTxState)
                pactape_write_tape();
            else if (state == st_demo)
                bCancelThisMode = 1;
        }
        if ((!bTapeRunning || bTxState) && bBeginLevel)
        {
            bTapeRunning = 1;
            moveCurrent = moves;
            nFrames = 0;
        }
    }
    else if (mode == mode_learn || mode == mode_best)
    {
        if (!bLearnEngaged)
        {
            if (bBeginLevel)
            {
                bLearnEngaged = 1;
                paclearn_init (Machine->gamedrv->name);
                goto learnReadyInit;
            }
        }
        else
        {
            if (state == st_ready)
            {
                if (bTxState)
                {
learnReadyInit:
                    xLast = 14;
                    yLast = 24;

                    if (bBeginLevel /* start || win */
                          && !paclearn_start (nLevel, mode == mode_best))
                        bCancelThisMode = 1;
                    else
                        maskChosen = (~paclearn_request (dL | dR)) & 0x0f;
                }
            }
            else if (state == st_play)
            {
                int dx, dy;

                if (bTxState)
                {
                    maskChosen = maskOverride;
                    paclearn_commit ((~maskOverride) & 0x0f);
                }
                /* 
                 * determine if a decision is necessary 
                 */

                xPacman = 239 - spriteram_2[nSpritePacman];
                yPacman = 272 - spriteram_2[nSpritePacman+1];

                x = xPacman/8 + 1;
                y = yPacman/8 - 1;
                dx = x-xLast;
                dy = y-yLast;

                if (dx || dy)
                {
                    static int bDeciding;
                    xLast = x;
                    yLast = y;
                    x += dx;
                    y += dy;

                    if (bDeciding)
                    {
                        maskChosen = maskOverride;
                        paclearn_commit ((~maskOverride) & 0x0f);
                        bDeciding = 0;
                    }
                    if (x > 0 && x < 27) /* ignore tunnel */
                    {
                        int dirOptions;
                        dirOptions = ((0xce > VID(x,y+1)) << 3)  /* D */
                                   | ((0xce > VID(x+1,y)) << 2)  /* R */
                                   | ((0xce > VID(x-1,y)) << 1)  /* L */
                                   | ((0xce > VID(x,y-1))     ); /* U */
                        if (!dirOptions) debug_msg ("NULL-MASK.");

                        /* Detect junction */
                        if ((dx && (dirOptions & (dU | dD))) ||
                            (dy && (dirOptions & (dL | dR))) )
                        {
                            /* Prevent reverse */
                                 if (dy > 0) dirOptions &= mU;
                            else if (dx > 0) dirOptions &= mL;
                            else if (dx < 0) dirOptions &= mR;
                            else             dirOptions &= mD;

                            maskChosen = (~paclearn_request (dirOptions)) & 0x0f;

                            bDeciding = 1;
                        }
                    }    
                }
            }
            else if (state == st_die || state == st_win)
            {
                if (bTxState)
                    paclearn_end (state == st_win);
            }
        }
    }

    if (bCancelThisMode)
    {
        if (mode == mode_record && bTapeRunning)
            pactape_write_tape();
        if (bLearnEngaged)
            paclearn_shutdown();
        bVersion1Hack = 0;
        bTapeRunning = 0;
        bLearnEngaged = 0;
        mode = mode_normal;
        bCancelThisMode = 0;
    }
}

/***************************************************************************

  Provide some visual feedback

***************************************************************************/
void pactape_add_display(struct osd_bitmap *bitmap)
{
#if 0 /* DEBUG: show sprite numbers */
    int i;
    for (i=2; i<14; i+=2)
    {
        int x = 239 - spriteram_2[i];
        int y = 272 - spriteram_2[i+1];
		drawgfx(bitmap,Machine->gfx[0],
			    i/2, 1,
			    0,0,x,y,
			    &Machine->drv->visible_area,TRANSPARENCY_NONE,0);
    }
#endif /* END DEBUG */

    /* Draw a letter in the upper-right corner according to mode
     * in this color:
     * 
     *      solid   BLUE = engaged, running                     
     *      solid    RED = cannot engage now                    
     *   flashing YELLOW = will start at beginning of next level
     */
    if (mode != mode_normal)
    {                                   /* 'L'   'B'   'R'   'A' */
        static const char ch [6] = {0, 0, 0x4c, 0x42, 0x52, 0x41};
        static int blink;
        int color;
        int bActive = (bTapeRunning || bLearnEngaged);

        if (bActive)             color = 0x05;
        else 
        {
            if (state == st_demo  ||
                state == st_start ||
                state == st_win   ||
                bBeginLevel       )
                                 color = (blink > 9) ? 0x07 : 0x00;
            else
            {
                if (bKeyPressed) color = 0x01;
                else           { color = 0x00; mode = mode_normal; }
            }
            blink = (blink + 29) % 30;
        }
		drawgfx(bitmap,Machine->gfx[0],
			    ch[mode], color,
			    0,0,8*27,0,
			    &Machine->drv->visible_area,TRANSPARENCY_NONE,0);
    }
    if (bLearnEngaged && mode == mode_learn && state == st_play)
    {
        static const char ch    [3] = {0x50, 0x45, 0x45};
        static const int  color [3] = {0x05, 0x09, 0x09};
		drawgfx(bitmap,Machine->gfx[0],
			    ch[modeLearn], color[modeLearn],
			    0,0,8*26,0,
			    &Machine->drv->visible_area,TRANSPARENCY_NONE,0);
    }
#ifdef PT_DEBUG
    pacdebug_show_state();
    if (bSingleStep)
    {
        while (!osd_key_pressed(OSD_KEY_SPACE));
        while ( osd_key_pressed(OSD_KEY_SPACE));
    }
#endif
}

/***************************************************************************

  These functions are called every time game input is desired.
  When the engine is engaged, it will read the user's inputs
  or override them and play the game itself.

***************************************************************************/
int pactape_input_port_1_r(int offset)
{
    int read = input_port_1_r(offset);
    if (state == st_start)
    {
        if (mode == mode_learn)
            read &= 0xdf; /* start 1 player */
        nLevel = 0; /* incremented to 1 during update_state */
    }
    return read;
}

int pactape_input_port_0_r(int offset)
{
    static int nHold = 15;
    static int maskTape = 0xffff;
    static int bEnd;

    int read = input_port_0_r(offset);
    if (mode == mode_learn)
    {
        if (state == st_demo)
        {
            if (nHold)
            {
                read &= 0xdf; /* coin insert */
                nHold--;
            }
        }
        else if (state == st_start)
            nHold = 15;
    }
    if (bLearnEngaged)
    {
        if (state == st_play || state == st_ready)
        {
            int maskPlayer = read & 0x0f;
            if ((0x0f != (maskPlayer | maskOptions)) && /* selected dir is possible */
                (0x0f == (maskPlayer | (maskPlayer+1)))) /* only one dir is selected */
              maskOverride = maskPlayer;
            else
              maskOverride = maskChosen;
            read = (read & 0xf0) | maskOverride;
        }
    }
    if (bTapeRunning)
    {
        switch (state)
        {
        case st_demo:
            bCancelThisMode = 1;
            break;
        case st_ready:
            bEnd = 0;
            if (mode == mode_record)
                *moveCurrent = ((read & 0x0f) << 12) | nFrames;
            else if (mode == mode_playback && (*moveCurrent >> 12))
            {
                maskTape = *moveCurrent >> 12;
                read = (read & 0xf0) | maskTape;
            }
            break;
        case st_die:
        case st_win:
            if (!bEnd)
            {
                bEnd = 1;
                moveCurrent++;
                nFrames = 0;
                if (mode == mode_record)
                {
                    *moveCurrent = maskTape | nFrames;
                    if (moveCurrent - moves >= MOVEBUF-2) /* leave room for final two entries */
                        bCancelThisMode = 1;
                }
            }
            break;
        case st_play:
            nFrames++;
            if (mode == mode_record)
            {
                maskTape = (read & 0x0f) << 12;
                if (nFrames > 0x0fff)
                    bCancelThisMode = 1;
                else
                {
                    if ((*moveCurrent & 0xf000) != maskTape)
                    {
                        moveCurrent++;
                        nFrames = 0;
                        if (moveCurrent - moves == MOVEBUF-2) /* leave room for final two entries */
                            bCancelThisMode = 1;
                    }
                    *moveCurrent = maskTape | nFrames;
                }
            }
            else if (mode == mode_playback)
            {
                if (nFrames > (*moveCurrent & 0x0fff))
                {
                    moveCurrent++;
                    nFrames = 0;
                    if ((moveCurrent - moves >= MOVEBUF-1)
                        || (*moveCurrent == 0))
                        bCancelThisMode = 1;
                    maskTape = *moveCurrent >> 12;
                }
                read = (read & 0xf0) | maskTape;
            }
            break;
        }
    }
    return read;
}

/***************************************************************************

  Disk file interface for "tapes"

***************************************************************************/
int pactape_read_tape()
{
    int bSuccess = 0;
    FILE *f;
    unsigned char name [100];
    sprintf (name, "%s.T%02d", Machine->gamedrv->name, nLevel);
    if ((f = fopen(name,"rb")) != 0)
    {
        int version;
        fread(name, 1, 12, f);
        if (sscanf (name, "PACTAPE%2d", &version) == 1
            && version <= PacTapeVersion)
        {
            unsigned short * move;

            if (version > 1)
            {
                fread(name, 1, 1, f);
                bSuccess = pactape_check_dsw(name[0]);
            }
            else
            {
                bVersion1Hack = 1;
                bSuccess = 1;
            }

            if (bSuccess)
            {
                move = moves;
                while (move-moves < MOVEBUF-1)
                {
                    if (fread(name, 1, 2, f) < 2) break;
                    *move  = name[0] << 8;
                    *move |= name[1];
                    move++;
                }
                *move = 0;
                bSuccess = 1;
            }
        }
        fclose(f);
    }
    if (!bSuccess)
        *moves = 0;
    return bSuccess;
}

int pactape_write_tape()
{
    int bSuccess = 0;
    FILE *f;
    unsigned char name [100];
    *(moveCurrent+1)=0;
    sprintf (name, "%s.T%02d", Machine->gamedrv->name, nLevel);
    if ((f = fopen(name,"wb")) != 0)
    {
        unsigned short * move;
        sprintf (name, "PACTAPE%02d", PacTapeVersion);
        fwrite(name, 1, 12, f);

        name[0] = pactape_get_dsw();
        fwrite(name, 1, 1, f);

        for (move = moves; *move && move <= moveCurrent; move++)
        {
            name[0] = (*move >> 8) & 0xff;
            name[1] = (*move     ) & 0xff;
            fwrite(name, 1, 2, f);
        }
        bSuccess = 1;
        fclose(f);
    }
    return bSuccess;
}


/***************************************************************************

  Helper functions

***************************************************************************/
long pactape_get_score()
{
    /* compute SCORE */
    long nScore = 0;
    int vidoff;
    unsigned char chVid;

    for (vidoff = 0x3fc; vidoff > 0x3f6; vidoff--)
    {
        nScore *= 10;
        chVid = videoram[vidoff];
        if (chVid<10) nScore += chVid;
    }
    return nScore;
}

int pactape_get_dsw()
{
    /* Don't mask this here -- allow files
       to remember ENTIRE DSW state.
       Mask it in the check, below. */
    return input_port_2_r(0);
}
int pactape_check_dsw (int dswCheck)
{
    struct DisplayText dt[3];
	struct NewInputPort *in;
    int port=0;

    if ((pactape_get_dsw() & dswMask) == (dswCheck & dswMask))
        return 1;

    dt[0].text = "Pac-Tape ALERT:\n\n"
                 "The current DIP SWITCH settings do not match "
                 "the settings used when data was recorded!\n\n"
                 "Press ENTER to change current settings to "
                 "match.  THIS WILL RESET THE MACHINE (like F3).\n\n"
                 "Press any other key to keep current settings."
                 ;
    dt[0].color = DT_COLOR_WHITE;
    dt[0].x = 8*0;
    dt[0].y = 8*4;
    dt[1].text = 0;
    displaytext(dt,1);
    if (osd_read_key() != OSD_KEY_ENTER)
        return 0;

	in = Machine->input_ports;
	while (in->type != IPT_END)
	{
		if ((port == 3) && /* DSW only */
            (in->type & ~IPF_MASK) == IPT_DIPSWITCH_NAME && 
			(in->type & IPF_UNUSED) == 0 &&
            (in->mask & dswMask))
		{
			in->default_value = dswCheck & in->mask;
		}
        else if ((in->type & ~IPF_MASK) == IPT_PORT) 
            port++;
		in++;
	}

    machine_reset();
    return 0;
}

void debug_msg (char * text)
{
#ifdef PT_DEBUG
    struct DisplayText dt[3];
    dt[0].text = text;
    dt[0].color = DT_COLOR_YELLOW;
    dt[0].x = 8*0;
    dt[0].y = 8*35;
    dt[1].text = 0;
#if 1
    displaytext(dt,0);
    while (!osd_key_pressed (OSD_KEY_SPACE) && !osd_key_pressed (OSD_KEY_ENTER));
    while (osd_key_pressed (OSD_KEY_SPACE));
    if (osd_key_pressed (OSD_KEY_ESC)) exit(1);
#endif
#endif /* PT_DEBUG */
    if (errorlog) fprintf (errorlog, "PACTAPE: %s\n", text);
}
