#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pacdefs.h"
#include "pactree.h"

/***************************************************************************
  LearnPath methods
***************************************************************************/

LearnPath LearnPath_new (int nSize)
{
    LearnPath obj = (LearnPath) malloc (sizeof(struct s_LearnPath));
    if (obj==0) debug_msg ("MALLOC LP");
    nSize++; /* leave room for 0 terminator */
    if (nSize == 0)
        nSize = cnMaxLearnMoves;
    obj->score = 0;
    obj->choices = dX;
    obj->moves = (LearnMove *) malloc (sizeof(LearnMove) * nSize);
    if (obj->moves==0) debug_msg ("MALLOC LM");
    obj->pos = obj->moves;
    *obj->moves = 0;
    obj->size = nSize;
    return obj;
}
void LearnPath_copy (LearnPath pathTarget, LearnPath pathSource)
{
    LearnMove * pos = pathSource->moves;
    pathTarget->pos = pathTarget->moves;
    pathTarget->score   = pathSource->score;
    pathTarget->choices = pathSource->choices;
    while (pos != pathSource->pos)
        *pathTarget->pos++ = *pos++;
    *pathTarget->pos = 0;
    if (pathTarget->pos - pathTarget->moves >= pathTarget->size)
    {
        char text[28];
        sprintf (text, "OVERRUN-COPY-%d.", pathTarget->size);
        debug_msg (text);
    }
}
void LearnPath_add (LearnPath path, LearnMove move)
{
    if (path->pos - path->moves >= path->size)
    {
        char text[28];
        sprintf (text, "OVERRUN-ADD-%d.", path->size);
        debug_msg (text);
    }
    else
    {
        *path->pos++ = move;
        *path->pos = 0;
    }
}
#ifndef PL_INLINE
void LearnPath_delete (LearnPath obj)
{
    if (obj->moves)
        free (obj->moves);
    free(obj);
}
void LearnPath_rewind (LearnPath path)
{
    path->pos = path->moves;
}
LearnMove LearnPath_play (LearnPath path)
{
    if (*path->pos == 0) debug_msg ("PLAY NULL");
    return *path->pos++;
}
int LearnPath_end_b (LearnPath path)
{
    return (*path->pos == 0);
}
int LearnPath_exhausted_b (LearnPath path)
{
    return (path->choices == 0);
}
void LearnPath_set_choices (LearnPath path, int choices)
{
    path->choices = choices;
}
int LearnPath_get_choices (LearnPath path)
{
    return path->choices;
}
int LearnPath_size (LearnPath path)
{
    return path->pos - path->moves;
}
void LearnPath_set_score (LearnPath path, unsigned long nPoints)
{
    path->score = 1000 + (nPoints/10) - LearnPath_size (path);
}
long LearnPath_get_score (LearnPath path)
{
    return path->score;
}
#endif

int LearnPath_write (LearnPath path, FILE *f)
{
    static unsigned char buf [4];
    int nSize;
    LearnMove * pos = path->moves;
    while (*pos++);
    nSize = (pos - path->moves) * sizeof (LearnMove);

    buf[0] = (nSize >> 8) & 0xff;
    buf[1] = (nSize     ) & 0xff;
    fwrite(buf, 1, 2, f);

    buf[0] = (path->score >> 24) & 0xff;
    buf[1] = (path->score >> 16) & 0xff;
    buf[2] = (path->score >>  8) & 0xff;
    buf[3] = (path->score      ) & 0xff;
    if (buf[0] & 0xf0)
        debug_msg ("SCORE-TOO-BIG");
    /* "choices" stored in high four bits of "score" */
    buf[0] |= path->choices << 4;
    fwrite(buf, 1, 4, f);

    return (nSize == fwrite(path->moves, 1, nSize, f));
}
LearnPath LearnPath_read (FILE *f)
{
    LearnPath path = pathNull;
    static unsigned char buf [4];
    int nSize;

    if (fread(buf, 1, 2, f) == 2)
    {
        nSize = buf[0] <<  8
              | buf[1];

        path = LearnPath_new (nSize/sizeof(LearnMove));

        fread(buf, 1, 4, f);
        path->score = (buf[0] & 0x0f) << 24
                    | buf[1] << 16
                    | buf[2] <<  8
                    | buf[3];
        /* "choices" stored in high four bits of "score" */
        path->choices = buf[0] >> 4;

        fread(path->moves, 1, nSize, f);
        path->moves[nSize/sizeof(LearnMove)] = 0;
    }
    else
        debug_msg ("READ-PATH-NULL.");
    return path;
}

/***************************************************************************
  LearnPathList methods
***************************************************************************/

LearnPathList LearnPathList_new()
{
    LearnPathList obj = (LearnPathList) malloc (sizeof(struct s_LearnPathList));
    if (obj==0) debug_msg ("MALLOC LPL");
    obj->next = obj->prev = 0;
    return obj;
}
#ifndef PL_INLINE
void LearnPathList_delete (LearnPathList obj)
{
    free(obj);
}
#endif

/***************************************************************************
  LearnTable methods
***************************************************************************/

LearnTable LearnTable_new()
{
    LearnTable obj = (LearnTable) malloc (sizeof(struct s_LearnTable));

    LearnPathList node = LearnPathList_new();
    if (obj==0) debug_msg ("MALLOC LT");
    node->path = LearnPath_new(0);
    node->path->choices = dR | dL;

    obj->head = obj->tail = obj->recent = node;
    obj->nNodes = 1;

    obj->data = 0;
    obj->nData = 0;

    return obj;
}
void LearnTable_delete (LearnTable obj)
{
    LearnPathList next = obj->head;
    LearnPathList node;
    while (next)
    {
        node = next;
        next = node->next;
        LearnPath_delete (node->path);
        LearnPathList_delete (node);
    }
    if (obj->nData)
        free(obj->data);
    free(obj);
}

void LearnTable_add (LearnTable table, LearnPath path)
{
    int score = LearnPath_get_score (path);
    int scoreR = LearnPath_get_score (table->recent->path);
    LearnPathList next, prev, node;

    if (!table->recent) debug_msg ("NO RECENT");
    if (scoreR > score)
    {
        next = table->recent;
        while (next && LearnPath_get_score (next->path) > score)
            next = next->next;
        if (next) prev = next->prev;
        else      prev = table->tail;
    }
    else
    {
        prev = table->recent;
        while (prev && LearnPath_get_score (prev->path) < score)
            prev = prev->prev;
        if (prev) next = prev->next;
        else      next = table->head;
    }
    node = LearnPathList_new();
    node->path = path;
    node->next = next;
    node->prev = prev;
    if (node->prev) node->prev->next = node;
    else                 table->head = node;
    if (node->next) node->next->prev = node;
    else                 table->tail = node;
    table->nNodes++;
    table->recent = node;

    LearnTable_assert_valid (table);
}

void LearnTable_remove (LearnTable table, LearnPath path)
{
    LearnPathList node = table->recent;
    if (!node) debug_msg ("RECENT NULL");

    while (node->path != path)
    {
        node = node->next;
        if (!node)
            node = table->head;
        if (node == table->recent)
            return;
    }
    if (node->prev) node->prev->next = node->next;
    else                 table->head = node->next;
    if (node->next) node->next->prev = node->prev;
    else                 table->tail = node->prev;
    if (table->recent == node)
        table->recent = (node->prev) ? node->prev : node->next;
    LearnPathList_delete(node);
    table->nNodes--;

    LearnTable_assert_valid (table);
}

#ifndef PL_INLINE

LearnPath LearnTable_best (LearnTable table)
{
    /* find best scoring path in table */
    table->recent = table->head;
    return table->recent->path;
}
void LearnTable_assert_valid (LearnTable table)
{
#ifdef PT_DEBUG
    LearnPathList prev, node;
    long count = 0;

    if (table->tail->next)
        debug_msg ("TAIL.");
    if (table->head->prev)
        debug_msg ("HEAD.");
    prev = 0;
    for (node = table->head; node; node = node->next)
    {
        count++;
        if (node->prev != prev)
        {
            sprintf (debugText[0], "IV:PREV-%ld.", count);
            debug_msg (debugText[0]);
        }
        if (prev && LearnPath_get_score (prev->path) < LearnPath_get_score (node->path))
        {
            sprintf (debugText[0], "IV:SCORE-%ld.", count);
            debug_msg (debugText[0]);
        }
        prev = node;
    }
    if (table->nNodes != count)
    {
        sprintf (debugText[0], "IV:NODES-%ld-%ld.", table->nNodes, count);
        debug_msg (debugText[0]);
    }
#endif
}
#endif

int LearnTable_write (LearnTable table, const char * name)
{
    int bSuccess = 0;
    FILE *f;
    unsigned char buf [16];
    if ((f = fopen(name,"wb")) != 0)
    {
        LearnPathList node;
        long nRecent = 0;
        long nCount = 0;

        LearnTable_assert_valid (table);

        sprintf (buf, "PACLEARN%02d", PacLearnVersion);
        fwrite(buf, 1, 12, f);

        buf[0] = (table->nData) & 0xff;
        fwrite(buf, 1, 1, f);
        if (table->nData)
            fwrite (table->data, 1, table->nData, f);

        buf[0] = (table->nNodes >> 24) & 0xff;
        buf[1] = (table->nNodes >> 16) & 0xff;
        buf[2] = (table->nNodes >>  8) & 0xff;
        buf[3] = (table->nNodes      ) & 0xff;
        fwrite(buf, 1, 4, f);
        for (node = table->head; node; node = node->next)
        {
            nCount++;
            if (node == table->recent)
                nRecent = nCount;
            if (!LearnPath_write (node->path, f))
                break;
        }
        if (!node)
        {
            buf[0] = (nRecent >> 24) & 0xff;
            buf[1] = (nRecent >> 16) & 0xff;
            buf[2] = (nRecent >>  8) & 0xff;
            buf[3] = (nRecent      ) & 0xff;
            fwrite(buf, 1, 4, f);
            bSuccess = 1;
        }
        fclose(f);
    }
    return bSuccess;
}

LearnTable LearnTable_read (const char * name)
{
    LearnTable table = tableNull;
    FILE *f;
    unsigned char buf [16];
    if ((f = fopen(name,"rb")) != 0)
    {
        LearnPathList prev = 0;
        LearnPathList node;
        LearnPath path = pathNull;
        long nRecent, nCount = 0;
        int version;

        fread(buf, 1, 12, f);
        if (sscanf (buf, "PACLEARN%2d", &version) == 1
            && version <= PacLearnVersion)
        {
            table = (LearnTable) malloc (sizeof(struct s_LearnTable));

            if (version > 1)
            {
                fread(buf, 1, 1, f);
                table->nData = buf[0];
                if (table->nData)
                {
                    table->data = (unsigned char *) malloc (table->nData);
                    fread (table->data, 1, table->nData, f);
                }
            }

            fread(buf, 1, 4, f);
            table->nNodes = buf[0] << 24
                          | buf[1] << 16
                          | buf[2] <<  8
                          | buf[3];
            nCount = table->nNodes;
            do 
            {
                path = LearnPath_read (f);
                if (path == pathNull)
                    break;
                node = LearnPathList_new();
                node->path = path;
                node->prev = prev;
                if (prev) prev->next = node;
                else      table->head = node;
                nCount--;
                prev = node;
            }
            while (nCount);
            table->tail = prev;

            fread(buf, 1, 4, f);
            nRecent = buf[0] << 24
                    | buf[1] << 16
                    | buf[2] <<  8
                    | buf[3];
            for (node = table->head; node; node = node->next)
            {
                nCount++;
                if (nRecent == nCount)
                {
                    table->recent = node;
                    break;
                }
            }
            LearnTable_assert_valid (table);
        }
        fclose(f);
    }
    return table;
}
