/***************************************************************************

  pactree.h

  Infrastructure for searching deterministic Pac-Man paths.
  Provides basic search-tree elements:
     LearnPath   - encoding of a (partial) path searched
     LearnTable  - ranked table of LearnPath elements

  Module added by Sean Gugler on 97-08-28

  modified 97-08-27 Sean Gugler - version 0.8

***************************************************************************/

#ifndef PACTREE_H
#define PACTREE_H

#define PacLearnVersion 2

/* Execution is faster with in-lined functions.          *
 * Define PL_INLINE to enable in-lining where possible.  */
#ifndef PT_DEBUG
#define PL_INLINE
#endif

/*******************************
      BASIC MOVEMENT UNIT
********************************/

/* dU, dL, dR, or dD (pacdefs.h) */
typedef unsigned char LearnMove;

/******************************************
  Path: scored chain of moves, with
        branches to generate longer chains
*******************************************/

#define cnMaxLearnMoves 500

typedef struct s_LearnPath
{
    LearnMove * moves;
    LearnMove * pos;
    long score;
    unsigned char choices;
    int size;
}
* LearnPath;

#define pathNull 0
#define pathPosNull 0

LearnPath LearnPath_new         (int nSize);
void      LearnPath_copy        (LearnPath pathTarget, LearnPath pathSource);
void      LearnPath_add         (LearnPath path,       LearnMove move);
#ifndef PL_INLINE /* IN-LINE-ABLE FUNCTIONS */
void      LearnPath_delete      (LearnPath obj);
void      LearnPath_rewind      (LearnPath path);
LearnMove LearnPath_play        (LearnPath path);
int       LearnPath_end_b       (LearnPath path);
int       LearnPath_exhausted_b (LearnPath path);
void      LearnPath_set_choices (LearnPath path, int choices);
int       LearnPath_get_choices (LearnPath path);
int       LearnPath_size        (LearnPath path);
void      LearnPath_set_score   (LearnPath path, unsigned long nPoints);
long      LearnPath_get_score   (LearnPath path);
#endif

/*******************************
    Doubly-linked list node
********************************/

typedef struct s_LearnPathList
{
    LearnPath path;
    struct s_LearnPathList * next;
    struct s_LearnPathList * prev;
}
* LearnPathList;

LearnPathList LearnPathList_new    ();
#ifndef PL_INLINE /* IN-LINE-ABLE FUNCTIONS */
void          LearnPathList_delete (LearnPathList obj);
#endif

/*******************************
     Search Table of paths
********************************/

typedef struct s_LearnTable
{
    LearnPathList head;
    LearnPathList tail;
    LearnPathList recent; /* Faster sorted insertion */
    long nNodes;
    unsigned char * data;
    int nData;
}
* LearnTable;

#define tableNull 0

LearnTable LearnTable_new          ();
void       LearnTable_delete       (LearnTable obj);
void       LearnTable_add          (LearnTable table, LearnPath path);
void       LearnTable_remove       (LearnTable table, LearnPath path);
int        LearnTable_write        (LearnTable table, const char * filename);
LearnTable LearnTable_read         (const char * filename);
#ifndef PL_INLINE /* IN-LINE-ABLE FUNCTIONS */
LearnPath  LearnTable_best         (LearnTable table);
void       LearnTable_assert_valid (LearnTable table);
#endif

/***************************************************************************

  IMPLEMENTATION of In-Lined functions

***************************************************************************/
#ifdef PL_INLINE

#define LearnPath_delete(obj) \
    if (obj->moves) free (obj->moves); free(obj)
#define LearnPath_rewind(path) \
    path->pos = path->moves
#define LearnPath_play(path) \
    (*path->pos++)
#define LearnPath_end_b(path) \
    (*path->pos == 0)
#define LearnPath_exhausted_b(path) \
    (path->choices == 0)
#define LearnPath_set_choices(path, c) \
    path->choices = c
#define LearnPath_get_choices(path) \
    (path->choices)
#define LearnPath_size(path) \
    (path->pos - path->moves)
#define LearnPath_set_score(path, nPoints) \
    path->score = 1000 + (nPoints/10) - LearnPath_size (path)
#define LearnPath_get_score(path) \
    (path->score)

#define LearnPathList_delete(obj) free(obj)

extern const char * textLearnFile;
#define LearnTable_set_file(name) \
    textLearnFile = name
#define LearnTable_best(table) \
    ((table->recent = table->head) ->path)
#define LearnTable_assert_valid(table) /* NOP */

#endif /* PL_INLINE */


#endif /* PACTREE_H */
